@@ -4,14 +4,87 @@ Conversions from this module operate on batches of poses or transformations
 and can be 400 to 1000 times faster than a loop of individual conversions.
 """
 import numpy as np
-from .plot_utils import Trajectory, make_3d_axis
-from .batch_rotations import (matrices_from_quaternions,
-                              quaternions_from_matrices,
-                              matrices_from_compact_axis_angles,
-                              axis_angles_from_matrices)
+from .batch_rotations import (
+    matrices_from_quaternions, quaternions_from_matrices,
+    matrices_from_compact_axis_angles, axis_angles_from_matrices,
+    batch_concatenate_quaternions, batch_q_conj)
 from .transformations import transform_from_exponential_coordinates
 
 
+def invert_transforms(A2Bs):
+    """Invert transforms.
+
+    Parameters
+    ----------
+    A2Bs : array-like, shape (..., 4, 4)
+        Transforms from frames A to frames B
+
+    Returns
+    -------
+    B2As : array, shape (..., 4, 4)
+        Transforms from frames B to frames A
+    """
+    A2Bs = np.asarray(A2Bs)
+    instances_shape = A2Bs.shape[:-2]
+    B2As = np.empty_like(A2Bs)
+    # ( R t )^-1   ( R^T -R^T*t )
+    # ( 0 1 )    = ( 0    1     )
+    B2As[..., :3, :3] = A2Bs[..., :3, :3].transpose(
+        list(range(A2Bs.ndim - 2)) + [A2Bs.ndim - 1, A2Bs.ndim - 2])
+    B2As[..., :3, 3] = np.einsum(
+        "nij,nj->ni",
+        -B2As[..., :3, :3].reshape(-1, 3, 3),
+        A2Bs[..., :3, 3].reshape(-1, 3)).reshape(
+        *(list(instances_shape) + [3]))
+    B2As[..., 3, :3] = 0.0
+    B2As[..., 3, 3] = 1.0
+    return B2As
+
+
+def concat_one_to_many(A2B, B2Cs):
+    """Concatenate transformation A2B with multiple transformations B2C.
+
+    We use the extrinsic convention, which means that B2Cs are left-multiplied
+    to A2B.
+
+    Parameters
+    ----------
+    A2B : array-like, shape (4, 4)
+        Transform from frame A to frame B
+
+    B2Cs : array-like, shape (n_transforms, 4, 4)
+        Transforms from frame B to frame C
+
+    Returns
+    -------
+    A2Cs : array, shape (n_transforms, 4, 4)
+        Transforms from frame A to frame C
+    """
+    return np.einsum("nij,jk->nik", B2Cs, A2B)
+
+
+def concat_many_to_one(A2Bs, B2C):
+    """Concatenate multiple transformations A2B with transformation B2C.
+
+    We use the extrinsic convention, which means that B2C is left-multiplied
+    to A2Bs.
+
+    Parameters
+    ----------
+    A2Bs : array-like, shape (4, 4)
+        Transforms from frame A to frame B
+
+    B2C : array-like, shape (n_transforms, 4, 4)
+        Transform from frame B to frame C
+
+    Returns
+    -------
+    A2Cs : array, shape (n_transforms, 4, 4)
+        Transforms from frame A to frame C
+    """
+    return np.einsum("ij,njk->nik", B2C, A2Bs)
+
+
 def transforms_from_pqs(P, normalize_quaternions=True):
     """Get sequence of homogeneous matrices from positions and quaternions.
 
@@ -108,21 +181,24 @@ def exponential_coordinates_from_transforms(A2Bs):
     thetas = omega_thetas[..., 3]
 
     # from sympy import *
-    # omega0, omega1, omega2, px, py, pz, theta = symbols("o0 o1 o2 p0 p1 p2 theta")
-    # w = Matrix([[0, -omega2, omega1], [omega2, 0, -omega0], [-omega1, omega0, 0]])
+    # o0, o1, o2, px, py, pz, t = symbols("o0 o1 o2 p0 p1 p2 theta")
+    # w = Matrix([[0, -o2, o1], [o2, 0, -o0], [-o1, o0, 0]])
     # p = Matrix([[px], [py], [pz]])
-    # v = (eye(3) / theta - 0.5 * w + (1.0 / theta - 0.5 / tan(theta / 2.0)) * w * w) * p
+    # v = (eye(3) / t - 0.5 * w + (1 / t - 0.5 / tan(t / 2.0)) * w * w) * p
 
     # Result:
-    # p0*(-o1**2*(-0.5/tan(0.5*theta) + 1.0/theta) - o2**2*(-0.5/tan(0.5*theta) + 1.0/theta) + 1/theta)
-    #     + p1*(o0*o1*(-0.5/tan(0.5*theta) + 1.0/theta) + 0.5*o2)
-    #     + p2*(o0*o2*(-0.5/tan(0.5*theta) + 1.0/theta) - 0.5*o1)
-    # p0*(o0*o1*(-0.5/tan(0.5*theta) + 1.0/theta) - 0.5*o2)
-    #     + p1*(-o0**2*(-0.5/tan(0.5*theta) + 1.0/theta) - o2**2*(-0.5/tan(0.5*theta) + 1.0/theta) + 1/theta)
-    #     + p2*(0.5*o0 + o1*o2*(-0.5/tan(0.5*theta) + 1.0/theta))
-    # p0*(o0*o2*(-0.5/tan(0.5*theta) + 1.0/theta) + 0.5*o1)
-    #     + p1*(-0.5*o0 + o1*o2*(-0.5/tan(0.5*theta) + 1.0/theta))
-    #     + p2*(-o0**2*(-0.5/tan(0.5*theta) + 1.0/theta) - o1**2*(-0.5/tan(0.5*theta) + 1.0/theta) + 1/theta)
+    # p0*(-o1**2*(-0.5/tan(0.5*t) + 1/t)
+    #     - o2**2*(-0.5/tan(0.5*t) + 1/t) + 1/t)
+    #     + p1*(o0*o1*(-0.5/tan(0.5*t) + 1/t) + 0.5*o2)
+    #     + p2*(o0*o2*(-0.5/tan(0.5*t) + 1/t) - 0.5*o1)
+    # p0*(o0*o1*(-0.5/tan(0.5*t) + 1/t) - 0.5*o2)
+    #     + p1*(-o0**2*(-0.5/tan(0.5*t) + 1/t)
+    #           - o2**2*(-0.5/tan(0.5*t) + 1/t) + 1/t)
+    #     + p2*(0.5*o0 + o1*o2*(-0.5/tan(0.5*t) + 1/t))
+    # p0*(o0*o2*(-0.5/tan(0.5*t) + 1/t) + 0.5*o1)
+    #     + p1*(-0.5*o0 + o1*o2*(-0.5/tan(0.5*t) + 1/t))
+    #     + p2*(-o0**2*(-0.5/tan(0.5*t) + 1/t)
+    #           - o1**2*(-0.5/tan(0.5*t) + 1/t) + 1/t)
 
     thetas = np.maximum(thetas, np.finfo(float).tiny)
     ti = 1.0 / thetas
@@ -199,21 +275,21 @@ def transforms_from_exponential_coordinates(Sthetas):
             axes=screw_axes[..., :3], angles=t, out=A2Bs[..., :3, :3])
 
         # from sympy import *
-        # omega0, omega1, omega2, vx, vy, vz, theta = symbols("omega_0 omega_1 omega_2 v_x v_y v_z theta")
-        # w = Matrix([[0, -omega2, omega1], [omega2, 0, -omega0], [-omega1, omega0, 0]])
+        # o0, o1, o2, vx, vy, vz, t = symbols("o0 o1 o2 v_x v_y v_z t")
+        # w = Matrix([[0, -o2, o1], [o2, 0, -o0], [-o1, o0, 0]])
         # v = Matrix([[vx], [vy], [vz]])
-        # p = (eye(3) * theta + (1 - cos(theta)) * w + (theta - sin(theta)) * w * w) * v
+        # p = (eye(3) * t + (1 - cos(t)) * w + (t - sin(t)) * w * w) * v
         #
         # Result:
-        # -v_x*(omega_1**2*(theta - sin(theta)) + omega_2**2*(theta - sin(theta)) - theta)
-        #     + v_y*(omega_0*omega_1*(theta - sin(theta)) + omega_2*(cos(theta) - 1))
-        #     + v_z*(omega_0*omega_2*(theta - sin(theta)) - omega_1*(cos(theta) - 1))
-        # v_x*(omega_0*omega_1*(theta - sin(theta)) - omega_2*(cos(theta) - 1))
-        #     - v_y*(omega_0**2*(theta - sin(theta)) + omega_2**2*(theta - sin(theta)) - theta)
-        #     + v_z*(omega_0*(cos(theta) - 1) + omega_1*omega_2*(theta - sin(theta)))
-        # v_x*(omega_0*omega_2*(theta - sin(theta)) + omega_1*(cos(theta) - 1))
-        #     - v_y*(omega_0*(cos(theta) - 1) - omega_1*omega_2*(theta - sin(theta)))
-        #     - v_z*(omega_0**2*(theta - sin(theta)) + omega_1**2*(theta - sin(theta)) - theta)
+        # -v_x*(o1**2*(t - sin(t)) + o2**2*(t - sin(t)) - t)
+        #     + v_y*(o0*o1*(t - sin(t)) + o2*(cos(t) - 1))
+        #     + v_z*(o0*o2*(t - sin(t)) - o1*(cos(t) - 1))
+        # v_x*(o0*o1*(t - sin(t)) - o2*(cos(t) - 1))
+        #     - v_y*(o0**2*(t - sin(t)) + o2**2*(t - sin(t)) - t)
+        #     + v_z*(o0*(cos(t) - 1) + o1*o2*(t - sin(t)))
+        # v_x*(o0*o2*(t - sin(t)) + o1*(cos(t) - 1))
+        #     - v_y*(o0*(cos(t) - 1) - o1*o2*(t - sin(t)))
+        #     - v_z*(o0**2*(t - sin(t)) + o1**2*(t - sin(t)) - t)
 
         tms = t - np.sin(t)
         cm1 = np.cos(t) - 1.0
@@ -251,7 +327,209 @@ def transforms_from_exponential_coordinates(Sthetas):
     return A2Bs
 
 
-def plot_trajectory(ax=None, P=None, normalize_quaternions=True, show_direction=True, n_frames=10, s=1.0, ax_s=1, **kwargs):
+def dual_quaternions_from_pqs(pqs):
+    """Get dual quaternions from positions and quaternions.
+
+    Parameters
+    ----------
+    pqs : array-like, shape (..., 7)
+        Poses represented by positions and quaternions in the
+        order (x, y, z, qw, qx, qy, qz)
+
+    Returns
+    -------
+    dqs : array, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+    """
+    pqs = np.asarray(pqs)
+    instances_shape = pqs.shape[:-1]
+    out = np.empty(list(instances_shape) + [8])
+
+    # orientation quaternion
+    out[..., :4] = pqs[..., 3:]
+
+    # use memory temporarily to store position
+    out[..., 4] = 0
+    out[..., 5:] = pqs[..., :3]
+
+    out[..., 4:] = 0.5 * batch_concatenate_quaternions(
+        out[..., 4:], out[..., :4])
+    return out
+
+
+def dual_quaternions_from_transforms(A2Bs):
+    """Get dual quaternions from transformations.
+
+    Parameters
+    ----------
+    A2Bs : array-like, shape (..., 4, 4)
+        Poses represented by homogeneous matrices
+
+    Returns
+    -------
+    dqs : array, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+    """
+    A2Bs = np.asarray(A2Bs)
+    instances_shape = A2Bs.shape[:-2]
+    out = np.empty(list(instances_shape) + [8])
+
+    # orientation quaternion
+    out[..., :4] = quaternions_from_matrices(A2Bs[..., :3, :3])
+
+    # use memory temporarily to store position
+    out[..., 4] = 0
+    out[..., 5:] = A2Bs[..., :3, 3]
+
+    out[..., 4:] = 0.5 * batch_concatenate_quaternions(
+        out[..., 4:], out[..., :4])
+    return out
+
+
+def pqs_from_dual_quaternions(dqs):
+    """Get positions and quaternions from dual quaternions.
+
+    Parameters
+    ----------
+    dqs : array-like, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+
+    Returns
+    -------
+    pqs : array, shape (..., 7)
+        Poses represented by positions and quaternions in the
+        order (x, y, z, qw, qx, qy, qz)
+    """
+    dqs = np.asarray(dqs)
+    instances_shape = dqs.shape[:-1]
+    out = np.empty(list(instances_shape) + [7])
+    out[..., 3:] = dqs[..., :4]
+    out[..., :3] = 2 * batch_concatenate_quaternions(
+        dqs[..., 4:], batch_q_conj(out[..., 3:]))[..., 1:]
+    return out
+
+
+def transforms_from_dual_quaternions(dqs):
+    """Get transformations from dual quaternions.
+
+    Parameters
+    ----------
+    dqs : array-like, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+
+    Returns
+    -------
+    A2Bs : array, shape (..., 4, 4)
+        Poses represented by homogeneous matrices
+    """
+    dqs = np.asarray(dqs)
+    instances_shape = dqs.shape[:-1]
+    out = np.empty(list(instances_shape) + [4, 4])
+    out[..., :3, :3] = matrices_from_quaternions(dqs[..., :4])
+    out[..., :3, 3] = 2 * batch_concatenate_quaternions(
+        dqs[..., 4:], batch_q_conj(dqs[..., :4]))[..., 1:]
+    out[..., 3, :3] = 0.0
+    out[..., 3, 3] = 1.0
+    return out
+
+
+def batch_dq_conj(dqs):
+    """Conjugate of dual quaternions.
+
+    There are three different conjugates for dual quaternions. The one that we
+    use here converts (pw, px, py, pz, qw, qx, qy, qz) to
+    (pw, -px, -py, -pz, -qw, qx, qy, qz). It is a combination of the quaternion
+    conjugate and the dual number conjugate.
+
+    Parameters
+    ----------
+    dqs : array-like, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+
+    Returns
+    -------
+    dq_conjugates : array-like, shape (..., 8)
+        Conjugates of dual quaternions: (pw, -px, -py, -pz, -qw, qx, qy, qz)
+    """
+    out = np.empty_like(dqs)
+    out[..., 0] = dqs[..., 0]
+    out[..., 1:5] = -dqs[..., 1:5]
+    out[..., 5:] = dqs[..., 5:]
+    return out
+
+
+def batch_concatenate_dual_quaternions(dqs1, dqs2):
+    """Concatenate dual quaternions.
+
+    Suppose we want to apply two extrinsic transforms given by dual
+    quaternions dq1 and dq2 to a vector v. We can either apply dq2 to v and
+    then dq1 to the result or we can concatenate dq1 and dq2 and apply the
+    result to v.
+
+    Parameters
+    ----------
+    dqs1 : array-like, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+
+    dqs2 : array-like, shape (..., 8)
+        Dual quaternions to represent transforms:
+        (pw, px, py, pz, qw, qx, qy, qz)
+
+    Returns
+    -------
+    dqs3 : array, shape (8,)
+        Products of the two batches of dual quaternions:
+        (pw, px, py, pz, qw, qx, qy, qz)
+    """
+    dqs1 = np.asarray(dqs1)
+    dqs2 = np.asarray(dqs2)
+
+    out = np.empty_like(dqs1)
+    out[..., :4] = batch_concatenate_quaternions(
+        dqs1[..., :4], dqs2[..., :4])
+    out[..., 4:] = (
+        batch_concatenate_quaternions(dqs1[..., :4], dqs2[..., 4:]) +
+        batch_concatenate_quaternions(dqs1[..., 4:], dqs2[..., :4]))
+    return out
+
+
+def batch_dq_prod_vector(dqs, V):
+    """Apply transforms represented by a dual quaternions to vectors.
+
+    Parameters
+    ----------
+    dqs : array-like, shape (..., 8)
+        Unit dual quaternions
+
+    V : array-like, shape (..., 3)
+        3d vectors
+
+    Returns
+    -------
+    W : array, shape (3,)
+        3d vectors
+    """
+    dqs = np.asarray(dqs)
+
+    v_dqs = np.empty_like(dqs)
+    v_dqs[..., 0] = 1.0
+    v_dqs[..., 1:5] = 0.0
+    v_dqs[..., 5:] = V
+    v_dq_transformed = batch_concatenate_dual_quaternions(
+        batch_concatenate_dual_quaternions(dqs, v_dqs),
+        batch_dq_conj(dqs))
+    return v_dq_transformed[5:]
+
+
+def plot_trajectory(
+        ax=None, P=None, normalize_quaternions=True, show_direction=True,
+        n_frames=10, s=1.0, ax_s=1, **kwargs):  # pragma: no cover
     """Plot pose trajectory.
 
     Parameters
@@ -285,14 +563,21 @@ def plot_trajectory(ax=None, P=None, normalize_quaternions=True, show_direction=
     -------
     ax : Matplotlib 3d axis
         New or old axis
+
+    Raises
+    ------
+    ValueError
+        If trajectory does not contain any elements.
     """
     if P is None or len(P) == 0:
         raise ValueError("Trajectory does not contain any elements.")
 
     if ax is None:
+        from .plot_utils import make_3d_axis
         ax = make_3d_axis(ax_s)
 
     A2Bs = transforms_from_pqs(P, normalize_quaternions)
+    from .plot_utils import Trajectory
     trajectory = Trajectory(A2Bs, show_direction, n_frames, s, **kwargs)
     trajectory.add_trajectory(ax)
 
