@@ -5,14 +5,13 @@ See :doc:`transform_manager` for more information.
 import numpy as np
 import scipy.sparse as sp
 from scipy.sparse import csgraph
-try:
+try:  # pragma: no cover
     import pydot
-    pydot_available = True
+    PYDOT_AVAILABLE = True
 except ImportError:
-    pydot_available = False
+    PYDOT_AVAILABLE = False
 from .transformations import (check_transform, invert_transform, concat,
                               plot_transform)
-from .plot_utils import make_3d_axis
 
 
 class TransformManager(object):
@@ -68,8 +67,8 @@ class TransformManager(object):
 
         # Result of shortest path algorithm:
         # distance matrix (distance is the number of transformations)
-        self.dist = None
-        self.predecessors = None
+        self.dist = np.empty(0)
+        self.predecessors = np.empty(0)
 
         self._cached_shortest_paths = {}
 
@@ -78,11 +77,11 @@ class TransformManager(object):
 
         Parameters
         ----------
-        from_frame : str
+        from_frame : Hashable
             Name of the frame for which the transformation is added in the
             to_frame coordinate system
 
-        to_frame : str
+        to_frame : Hashable
             Name of the frame in which the transformation is defined
 
         A2B : array-like, shape (4, 4)
@@ -125,11 +124,11 @@ class TransformManager(object):
 
         Parameters
         ----------
-        from_frame : str
+        from_frame : Hashable
             Name of the frame for which the transformation is added in the
             to_frame coordinate system
 
-        to_frame : str
+        to_frame : Hashable
             Name of the frame in which the transformation is defined
 
         Returns
@@ -162,7 +161,7 @@ class TransformManager(object):
 
         Parameters
         ----------
-        frame : str
+        frame : Hashable
             Frame name
 
         Returns
@@ -177,11 +176,11 @@ class TransformManager(object):
 
         Parameters
         ----------
-        from_frame : str
+        from_frame : Hashable
             Name of the frame for which the transformation is requested in the
             to_frame coordinate system
 
-        to_frame : str
+        to_frame : Hashable
             Name of the frame in which the transformation is defined
 
         Returns
@@ -189,6 +188,12 @@ class TransformManager(object):
         A2B : array-like, shape (4, 4)
             Homogeneous matrix that represents the transformation from
             'from_frame' to 'to_frame'
+
+        Raises
+        ------
+        KeyError
+            If one of the frames is unknown or there is no connection between
+            them
         """
         if self.check:
             if from_frame not in self.nodes:
@@ -231,7 +236,8 @@ class TransformManager(object):
                          strict_check=self.strict_check, check=self.check)
         return A2B
 
-    def plot_frames_in(self, frame, ax=None, s=1.0, ax_s=1, show_name=True, whitelist=None, **kwargs):
+    def plot_frames_in(self, frame, ax=None, s=1.0, ax_s=1, show_name=True,
+                       whitelist=None, **kwargs):  # pragma: no cover
         """Plot all frames in a given reference frame.
 
         Note that frames that cannot be connected to the reference frame are
@@ -239,7 +245,7 @@ class TransformManager(object):
 
         Parameters
         ----------
-        frame : str
+        frame : Hashable
             Reference frame
 
         ax : Matplotlib 3d axis, optional (default: None)
@@ -264,6 +270,11 @@ class TransformManager(object):
         -------
         ax : Matplotlib 3d axis
             New or old axis
+
+        Raises
+        ------
+        KeyError
+            If the frame is unknown
         """
         if frame not in self.nodes:
             raise KeyError("Unknown frame '%s'" % frame)
@@ -281,7 +292,8 @@ class TransformManager(object):
                 pass  # Frame is not connected to the reference frame
         return ax
 
-    def plot_connections_in(self, frame, ax=None, ax_s=1, whitelist=None, **kwargs):
+    def plot_connections_in(self, frame, ax=None, ax_s=1, whitelist=None,
+                            **kwargs):  # pragma: no cover
         """Plot direct frame connections in a given reference frame.
 
         A line between each pair of frames for which a direct transformation
@@ -293,7 +305,7 @@ class TransformManager(object):
 
         Parameters
         ----------
-        frame : str
+        frame : Hashable
             Reference frame
 
         ax : Matplotlib 3d axis, optional (default: None)
@@ -313,11 +325,17 @@ class TransformManager(object):
         -------
         ax : Matplotlib 3d axis
             New or old axis
+
+        Raises
+        ------
+        KeyError
+            If the frame is unknown
         """
         if frame not in self.nodes:
             raise KeyError("Unknown frame '%s'" % frame)
 
         if ax is None:
+            from .plot_utils import make_3d_axis
             ax = make_3d_axis(ax_s)
 
         nodes = self._whitelisted_nodes(whitelist)
@@ -325,7 +343,7 @@ class TransformManager(object):
         if "c" not in kwargs and "color" not in kwargs:
             kwargs["color"] = "black"
 
-        for frame_names, transform in self.transforms.items():
+        for frame_names in self.transforms:
             from_frame, to_frame = frame_names
             if from_frame in nodes and to_frame in nodes:
                 try:
@@ -345,8 +363,6 @@ class TransformManager(object):
     def _whitelisted_nodes(self, whitelist):
         """Get whitelisted nodes.
 
-        A KeyError will be raised if an unknown node is in the whitelist.
-
         Parameters
         ----------
         whitelist : list or None
@@ -356,6 +372,11 @@ class TransformManager(object):
         -------
         nodes : set
             Existing whitelisted nodes
+
+        Raises
+        ------
+        KeyError
+            Will be raised if an unknown node is in the whitelist.
         """
         nodes = set(self.nodes)
         if whitelist is not None:
@@ -382,16 +403,16 @@ class TransformManager(object):
             of B2A?
         """
         consistent = True
-        for n1 in self.nodes:
-            for n2 in self.nodes:
+        for node1 in self.nodes:
+            for node2 in self.nodes:
                 try:
-                    n1_to_n2 = self.get_transform(n1, n2)
-                    n2_to_n1 = self.get_transform(n2, n1)
-                    n1_to_n2_inv = invert_transform(
-                        n2_to_n1, strict_check=self.strict_check,
+                    node1_to_node2 = self.get_transform(node1, node2)
+                    node2_to_node1 = self.get_transform(node2, node1)
+                    node1_to_node2_inv = invert_transform(
+                        node2_to_node1, strict_check=self.strict_check,
                         check=self.check)
-                    consistent = (consistent and
-                                  np.allclose(n1_to_n2, n1_to_n2_inv))
+                    consistent = consistent and np.allclose(node1_to_node2,
+                                                            node1_to_node2_inv)
                 except KeyError:
                     pass  # Frames are not connected
         return consistent
@@ -410,7 +431,7 @@ class TransformManager(object):
         return csgraph.connected_components(
             self.connections, directed=False, return_labels=False)
 
-    def write_png(self, filename, prog=None):
+    def write_png(self, filename, prog=None):  # pragma: no cover
         """Create PNG from dot graph of the transformations.
 
         .. warning::
@@ -427,8 +448,13 @@ class TransformManager(object):
             Name of GraphViz executable that can be found in the `$PATH` or
             absolute path to GraphViz executable. Possible options are, for
             example, 'dot', 'twopi', 'neato', 'circo', 'fdp', 'sfdp'.
+
+        Raises
+        ------
+        ImportError
+            If pydot is not available
         """
-        if not pydot_available:
+        if not PYDOT_AVAILABLE:
             raise ImportError("pydot must be installed to use this feature.")
 
         graph = pydot.Dot(graph_type="graph")
@@ -437,26 +463,27 @@ class TransformManager(object):
 
         for frame in self.nodes:
             node = pydot.Node(
-                self.__display_name(frame), style="filled",
+                _dot_display_name(frame), style="filled",
                 fillcolor=frame_color, shape="egg")
             graph.add_node(node)
         for frames, A2B in self.transforms.items():
-            a, b = frames
+            frame_a, frame_b = frames
             connection_name = "%s to %s\n%s" % (
-                self.__display_name(a), self.__display_name(b),
+                _dot_display_name(frame_a), _dot_display_name(frame_b),
                 str(np.round(A2B, 3)))
             node = pydot.Node(
                 connection_name, style="filled", fillcolor=connection_color,
                 shape="note")
             graph.add_node(node)
-            a_name = self.__display_name(a)
+            a_name = _dot_display_name(frame_a)
             a_edge = pydot.Edge(connection_name, a_name, penwidth=3)
             graph.add_edge(a_edge)
-            b_name = self.__display_name(b)
+            b_name = _dot_display_name(frame_b)
             b_edge = pydot.Edge(connection_name, b_name, penwidth=3)
             graph.add_edge(b_edge)
 
         graph.write_png(filename, prog=prog)
 
-    def __display_name(self, name):
-        return name.replace("/", "")
+
+def _dot_display_name(name):  # pragma: no cover
+    return name.replace("/", "")
